
#include "set.h"
#include "config.h"


/**
 * Allocate memory and set zero a new mySet struct.
 * It does not allocate its members.
 * @return Pointer of mySet type
 */
mySet *set_new()
{
	mySet *returnValue;

	returnValue = (mySet *) malloc(sizeof(mySet));
	memset(returnValue, 0, sizeof(mySet));

	return returnValue;
}

/**
 * Free the memory of element(int array) of a mySet struct
 * if mySet is not empty. It does not free mySet itself. 
 * @param aSet Pointer of mySet type
 */
void set_free(mySet *aSet)
{
	if (aSet->noOfElt > 0)
	{
		free(aSet->elt);
	}
}

/**
 * Free the memory of all members of a mySetSet struct
 * recursively. It does not free mySetSet itself.
 * @param aSetSet Pointer of type mySetSet
 */
void setset_free(mySetSet *aSetSet)
{
	int i;
	
	if (aSetSet->noOfElt > 0)
	{
		for (i = 0; i < aSetSet->noOfElt; i++)
		{
			set_free(aSetSet->elt[i]);
			free(aSetSet->elt[i]);
		}

		free(aSetSet->elt);
	}
}

/**
 * It output all members(int) of struct mySet to a file.
 * @param filePtr Pointer to a opened file.
 * @param aSet Pointer of struct mySet
 */
void set_print(FILE *filePtr, mySet *aSet)
{
	int i;

	for (i = 0; i < aSet->noOfElt; i++)
	{
		fprintf(filePtr, "%d ", aSet->elt[i]);
	}
}

/**
 * It insert a new member(int) to struct mySet without
 * checking for duplication.
 * @param aSet Pointer to struct mySet
 * @param no new member to be added
 */
void set_insert(mySet *aSet, int no)
{
	int noOfElt;

	noOfElt = aSet->noOfElt;

	if (noOfElt == 0)
	{
		aSet->elt = (int *) malloc(sizeof(int));
	}
	else
	{
		aSet->elt = (int *) realloc(aSet->elt, sizeof(int)*(noOfElt+1));
	}
	aSet->elt[noOfElt] = no;

	aSet->noOfElt++;
}

/**
 * It insert a new member(int) to struct mySet. It maintain set
 * characteristic as it ignore duplication
 * @param aSet Pointer to struct mySet
 * @param no new member to be added
 */
void set_insertNonRedundant(mySet *aSet, int no)
{
	int noOfElt;
	int i;

	int *elt;

	int foundIndex;

	noOfElt = aSet->noOfElt;
	elt = aSet->elt;


	if (noOfElt == 0)
	{
		aSet->elt = (int *) malloc(sizeof(int));
		aSet->elt[noOfElt] = no;
	}
	else
	{
		foundIndex = -1;
		for (i = 0; i < noOfElt; i++)
		{
			if (no == elt[i])
			{
				foundIndex = i;
				break;
			}
		}

		if (foundIndex == -1)
		{
			aSet->elt = (int *) realloc(aSet->elt, sizeof(int)*(noOfElt+1));
			aSet->elt[noOfElt] = no;
		}
		else
		{
			// do noting (because it exists)
		}
	}
	
	aSet->noOfElt++;
}

/**
 * It copy and overwrite all members from a mySet struct to
 * anothers.
 * @param set1 Pointer to destination mySet. It should be
 *             allocated beforehand.
 * @param set2 Pointer to source mySet struct.
 */
void set_copy(mySet *set1, mySet *set2)
{
	if (set1->noOfElt > 0)
	{
		free(set1->elt);
	}

	set1->noOfElt = set2->noOfElt;
	set1->elt = (int *) malloc(sizeof(int)*set1->noOfElt);
	memcpy(set1->elt, set2->elt, sizeof(int)*set1->noOfElt);
}

/**
 * It delete the ith(index) element in mySet struct.
 * Assumption: index is within range of the set 
 * i.e. 0 <= index < noOfElt
 * @param aSet Poitner to mySet struct
 * @param index Indicate which element to delete
 */
void set_remove(mySet *aSet, int index)
{
	int noOfElt;

	int *newElt;
	int *oldElt;

	noOfElt = aSet->noOfElt;
	oldElt = aSet->elt;

	if (noOfElt == 0)
	{
		return;
	}

	if (noOfElt == 1)
	{
		free(oldElt);
	}
	else
	{
		newElt = (int *) malloc(sizeof(int)*(noOfElt-1));

		if (index >= 1)
		{
			memcpy(newElt, oldElt, sizeof(int)*index);
		}

		if (index < noOfElt-1)
		{
			memcpy(&(newElt[index]), &(oldElt[index+1]), sizeof(int)*(noOfElt-index-1));
		}

		aSet->elt = newElt;
		free(oldElt);
	}
	aSet->noOfElt--;
}
 
/**
 * It add all members of a mySet struct to another mySet struct
 * without checking for duplication. Source mySet is left
 * untouched.
 * @param set1 Pointer to destination mySet struct. Member will
 *             be added to it.
 * @param set2 Pointer to source mySet struct. Left unchanged.
 */
void set_union(mySet *set1, mySet *set2)
{
	int noOfElt1, noOfElt2;
	int *elt1;
	int *elt2;

	int finalNoOfElt;
	int *finalElt;

	noOfElt1 = set1->noOfElt;
	noOfElt2 = set2->noOfElt;

	elt1 = set1->elt;
	elt2 = set2->elt;

	finalNoOfElt = noOfElt1+noOfElt2;

	finalElt = (int *) malloc(sizeof(int)*finalNoOfElt);
	if (noOfElt1 > 0)
	{
		memcpy(finalElt, elt1, sizeof(int)*noOfElt1);
	}
	
	if (noOfElt2 > 0)
	{
		memcpy(&(finalElt[noOfElt1]), elt2, sizeof(int)*noOfElt2);
	}

	free(elt1);

	set1->elt = finalElt;
	set1->noOfElt = finalNoOfElt;
}

/**
 * It check if a given integer is a member of a mySet struct.
 * @param aSet Pointer to mySet struct to look into.
 * @param eltNo Target value to look up.
 * @return 1 if it is a member. 0 if it is not a member.
 */
int set_isContain(mySet *aSet, int eltNo)
{
	int returnValue;
	int *eltList;

	int i;

	returnValue = 0;
	eltList = aSet->elt;
	for (i = 0; i < aSet->noOfElt; i++)
	{
		if (eltList[i] == eltNo)
		{
			returnValue = 1;
			break;
		}
	}

	return returnValue;
}

/**
 * It perform set union of two mySet struct. The final set
 * will be pointed by set1. set2 will left untouched. It
 * will check for duplication so that the result remain a set.
 * @param set1 Pointer to destination mySet struct. Member will
 *             be added to it.
 * @param set2 Pointer to source mySet struct. Left unchanged.
 */
void set_unionNonRedundant(mySet *set1, mySet *set2)
{
	int noOfElt1, noOfElt2;
	int *elt1;
	int *elt2;

	int finalNoOfElt;
	int *finalElt;

	int *isExist;

	int noOfAdd;

	int i;

	int eltNo;

	int tempNo;

	noOfElt1 = set1->noOfElt;
	noOfElt2 = set2->noOfElt;

	elt1 = set1->elt;
	elt2 = set2->elt;

	isExist = (int *) malloc(sizeof(int)*noOfElt2);
	memset(isExist, 0, sizeof(int)*noOfElt2);

	for (i = 0; i < noOfElt2; i++)
	{
		eltNo = elt2[i];
		isExist[i] = set_isContain(set1, eltNo);
	}
	noOfAdd = 0;
	for (i = 0; i < noOfElt2; i++)
	{
		if (isExist[i] == 0)
		{
			noOfAdd++;
		}
	}

	finalNoOfElt = noOfElt1+noOfAdd;

	finalElt = (int *) malloc(sizeof(int)*finalNoOfElt);
	if (noOfElt1 > 0)
	{
		memcpy(finalElt, elt1, sizeof(int)*noOfElt1);
	}

	tempNo = 0;
	for (i = 0; i < noOfElt2; i++)
	{
		if (isExist[i] == 0)
		{
			eltNo = elt2[i];

			finalElt[tempNo+noOfElt1] = eltNo;
			tempNo++;
		}
	}

	free(isExist);

	free(elt1);

	set1->elt = finalElt;
	set1->noOfElt = finalNoOfElt;

}

/**
 * It perform set complementation set1 - set2. The result will be kept
 * at set1. set2 will remain unchanged.
 * @param set1 Pointer to mySet struct
 * @param set2 Pointer to mySet struct
 */
void set_minus(mySet *set1, mySet *set2)
{
	int noOfElt1, noOfElt2;

	int i, j;

	int *elt1;
	int *elt2;

	int eltNo1, eltNo2;

	int *isMark;

	noOfElt1 = set1->noOfElt;
	noOfElt2 = set2->noOfElt;

	elt1 = set1->elt;
	elt2 = set2->elt;

	if (noOfElt1 > 0)
	{

		isMark = (int *) malloc(sizeof(int)*noOfElt1);
		memset(isMark, 0, sizeof(int)*noOfElt1);

		// mark delete
		for (i = 0; i < noOfElt1; i++)
		{
			eltNo1 = elt1[i];
			for (j = 0; j < noOfElt2; j++)
			{
				eltNo2 = elt2[j];
				if (eltNo1 == eltNo2)
				{
					isMark[i] = 1;
					break;
				}
			}
		}

		// delte
		for (i = noOfElt1-1; i >= 0; i--)
		{
			if (isMark[i] == 1)
			{
				set_remove(set1, i);
			}
		}

		free(isMark);
	}
}


/**
 * It perform intersection of set1 and set2. The result is
 * stored in set1. set2 is left untouched.
 * @param set1 Pointer to mySet struct
 * @param set2 Pointer to mySet struct
 */
void set_intersection(mySet *set1, mySet *set2)
{
	int noOfElt1, noOfElt2;
	int *elt1;
	int *elt2;

	int finalNoOfElt;

	int *finalElt;

	int *isMark;

	int eltNo1, eltNo2;

	int i, j;

	int tempNo;

	elt1 = set1->elt;
	noOfElt1 = set1->noOfElt;

	elt2 = set2->elt;
	noOfElt2 = set2->noOfElt;

	if (noOfElt1 > 0)
	{
		isMark = (int *) malloc(sizeof(int)*noOfElt1);
		memset(isMark, 0, sizeof(int)*noOfElt1);

		finalNoOfElt = 0;
		for (i = 0; i < noOfElt1; i++)
		{
			eltNo1 = elt1[i];
			for (j = 0; j < noOfElt2; j++)
			{
				eltNo2 = elt2[j];

				if (eltNo1 == eltNo2)
				{
					isMark[i] = 1;
					finalNoOfElt++;
					break;
				}
			}
		}

		if (finalNoOfElt > 0)
		{
			tempNo = 0;
			finalElt = (int *) malloc(sizeof(int)*finalNoOfElt);
			for (i = 0; i < noOfElt1; i++)
			{
				if (isMark[i] == 1)
				{
					eltNo1 = elt1[i];

					finalElt[tempNo] = eltNo1;
					tempNo++;
				}
			}
		}
		else
		{
			finalElt = NULL;
		}

		if (noOfElt1 > 0)
		{
			free(elt1);
		}

		set1->elt = finalElt;
		set1->noOfElt = finalNoOfElt;

		free(isMark);
	}
}

/**
 * <b>Deprediated:</b>It is not called by any function.
 * Perform intersection of set1 and set2 return the result to set1
 */
void set_intersection_class(mySet *set1, mySet *set2, mySet *set1Class, mySet *set2Class, int noOfClass, mySet *intersectClass)
{
	int noOfElt1, noOfElt2;
	int *elt1;
	int *elt2;

	int finalNoOfElt;

	int *finalElt;

	int *isMark;
	int *isMark2;

	int eltNo1, eltNo2;

	int i, j, k;

	int tempNo;

	int eltNoClass1, eltNoClass2;

	int *eltClass1;
	int *eltClass2;

	int isExistBoth;

	int isContain;

	elt1 = set1->elt;
	noOfElt1 = set1->noOfElt;

	elt2 = set2->elt;
	noOfElt2 = set2->noOfElt;

	eltClass1 = set1Class->elt;
	eltClass2 = set2Class->elt;

	if (noOfElt1 > 0)
	{
		isMark = (int *) malloc(sizeof(int)*noOfElt1);
		memset(isMark, 0, sizeof(int)*noOfElt1);

		isMark2 = (int *) malloc(sizeof(int)*noOfElt2);
		memset(isMark2, 0, sizeof(int)*noOfElt2);

		finalNoOfElt = 0;

		for (k = 0; k < noOfClass; k++)
		{
			isContain = set_isContain(intersectClass, k);

			isExistBoth = 0;
			for (i = 0; i < noOfElt1; i++)
			{
				eltNo1 = elt1[i];
				eltNoClass1 = eltClass1[i];

				if (eltNoClass1 == k)
				{
					for (j = 0; j < noOfElt2; j++)
					{
						eltNo2 = elt2[j];
						eltNoClass2 = eltClass2[j];

						if (eltNoClass2 == k)
						{
							isExistBoth = 1;
							if (eltNo1 == eltNo2)
							{
								isMark[i] = 1;
								finalNoOfElt++;
								break;
							}
						}
					}
				}
			}

			
			if ((isExistBoth == 0) && (isContain == 0))
			{
				for (i = 0; i < noOfElt1; i++)
				{
					eltNo1 = elt1[i];
					eltNoClass1 = eltClass1[i];

					if (eltNoClass1 == k)
					{
						isMark[i] = 1;
						finalNoOfElt++;
					}
				}
			}

			// reverse
			isExistBoth = 0;
			for (i = 0; i < noOfElt2; i++)
			{
				eltNo2 = elt2[i];
				eltNoClass2 = eltClass2[i];

				if (eltNoClass2 == k)
				{
					for (j = 0; j < noOfElt1; j++)
					{
						eltNo1 = elt1[j];
						eltNoClass1 = eltClass1[j];

						if (eltNoClass1 == k)
						{
							isExistBoth = 1;
							if (eltNo1 == eltNo2)
							{
								//isMark2[i] = 1;
								//finalNoOfElt++;
								break;
							}
						}
					}
				}
			}

			if ((isExistBoth == 0) && (isContain == 0))
			{
				for (i = 0; i < noOfElt2; i++)
				{
					eltNo2 = elt2[i];
					eltNoClass2 = eltClass2[i];

					if (eltNoClass2 == k)
					{
						isMark2[i] = 1;
						finalNoOfElt++;
					}
				}
			}
		}

		if (finalNoOfElt > 0)
		{
			tempNo = 0;
			finalElt = (int *) malloc(sizeof(int)*finalNoOfElt);
			for (i = 0; i < noOfElt1; i++)
			{
				if (isMark[i] == 1)
				{
					eltNo1 = elt1[i];

					finalElt[tempNo] = eltNo1;
					tempNo++;
				}
			}

			for (i = 0; i < noOfElt2; i++)
			{
				if (isMark2[i] == 1)
				{
					eltNo2 = elt2[i];

					finalElt[tempNo] = eltNo2;
					tempNo++;
				}
			}
		}
		else
		{
			finalElt = NULL;
		}

		if (noOfElt1 > 0)
		{
			free(elt1);
		}

		set1->elt = finalElt;
		set1->noOfElt = finalNoOfElt;

		free(isMark);
	}
}

/**
 * <b>Depreciated:</b>It is not called by any function
 */
void set_intersection_classSet(config *aConfig, mySet *set1, mySet *set2, myClassSet *set1Class, myClassSet *set2Class, myClassSet *intersectClass)
{
	int noOfElt1, noOfElt2;

	int *elt1;
	int *elt2;

	int i, j;

	int eltNo1, eltNo2;

	int noOfNumAttr;

	int *isMark1;
	int *isMark2;
	int *isMark3;

	//myClassSet *aClassSet1;
	//myClassSet *aClassSet2;

	int *finalElt;
	int tempNo;

	int finalNoOfElt;

	int dimNo, classID;

	int isContain;

	int count;

	int isExist;

	int *finalElt2;

	//int elt1, elt2;

	noOfElt1 = set1->noOfElt;
	noOfElt2 = set2->noOfElt;

	elt1 = set1->elt;
	elt2 = set2->elt;

	noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;

	isMark1 = (int *) malloc(sizeof(int)*noOfElt1);
	isMark2 = (int *) malloc(sizeof(int)*noOfElt2);

	memset(isMark1, 0, sizeof(int)*noOfElt1);
	memset(isMark2, 0, sizeof(int)*noOfElt2);
	
	finalNoOfElt= 0;
	// check not in intersectionClass
	for (i = 0; i < noOfElt1; i++)
	{
		eltNo1 = elt1[i];

		for (j = 0; j < aConfig->noOfCat; j++)
		{
			dimNo = noOfNumAttr+j;
			classID = aConfig->data[eltNo1].content[dimNo];

			isContain = classSet_isContain(intersectClass, dimNo, classID);
			if (isContain == 0)
			{
				isMark1[i] = 1;
				finalNoOfElt++;
				break;
			}
		}
	}
	for (i = 0; i < noOfElt2; i++)
	{
		eltNo2 = elt2[i];
		
		for (j = 0; j < aConfig->noOfCat; j++)
		{
			dimNo = noOfNumAttr+j;
			classID = aConfig->data[eltNo2].content[dimNo];

			isContain = classSet_isContain(intersectClass, dimNo, classID);
			if (isContain == 0)
			{
				isMark2[i] = 1;
				finalNoOfElt++;
				break;
			}
		}
	}

	// not intersect
	for (i = 0; i < noOfElt1; i++)
	{
		if (isMark1[i] == 0)
		{
			for (j = 0; j < noOfElt2; j++)
			{
				if (isMark2[j] == 0)
				{
					if (elt1[i] == elt2[j])
					{
						isMark1[i] = 1;
						finalNoOfElt++;
						break;
					}
				}
			}
		}
	}

	if (finalNoOfElt > 0)
	{
		tempNo = 0;
		finalElt = (int *) malloc(sizeof(int)*finalNoOfElt);
		for (i = 0; i < noOfElt1; i++)
		{
			if (isMark1[i] == 1)
			{
				finalElt[tempNo] = elt1[i];
				tempNo++;
			}
		}
		for (i = 0; i < noOfElt2; i++)
		{
			if (isMark2[i] == 1)
			{
				finalElt[tempNo] = elt2[i];
				tempNo++;
			}
		}


		count = 0;
		isMark3 = (int *) malloc(sizeof(int)*finalNoOfElt);
		memset(isMark3, 0, sizeof(int)*finalNoOfElt);
		for (i = 0; i < finalNoOfElt; i++)
		{
			isExist = 0;
			for (j = i+1; j < finalNoOfElt; j++)
			{
				if (finalElt[i] == finalElt[j])
				{
					isExist = 1;
					if (isMark3[i] == 0)
					{
						isMark3[i] = 1;
						count++;
					}
					if (isMark3[j] == 0)
					{
						isMark3[j] = 2;
					}
					//break;
				}
				
			}

			if (isExist == 0)
			{
				if (isMark3[i] == 0)
				{
					isMark3[i] = 1;
					count++;
				}
			}
		}


		finalElt2 = (int *) malloc(sizeof(int)*count);

		tempNo = 0;
		for (i = 0; i < finalNoOfElt; i++)
		{
			if (isMark3[i] == 1)
			{
				finalElt2[tempNo] = finalElt[i];
				tempNo++;
			}
		}

		free(finalElt);

		finalElt = finalElt2;
		finalNoOfElt = count;

		free(isMark3);
	}
	else
	{
		finalElt = NULL;
	}



	if (noOfElt1 > 0)
	{
		free(elt1);
	}

	set1->elt = finalElt;
	set1->noOfElt = finalNoOfElt;

	free(isMark1);
	free(isMark2);
}

/**
 * It allocate a new mySetSet struct and initialise it to zero.
 * @return Pointer to mySetSet struct
 */
mySetSet *setset_new()
{
	mySetSet *returnValue;

	returnValue = (mySetSet *) malloc(sizeof(mySetSet));
	memset(returnValue, 0, sizeof(mySetSet));

	return returnValue;
}

/**
 * It insert a mySet struct to a mySetSet struct without checking
 * duplication. Number of member of the mySetSet struct is increased by one
 * @param aSetSet Pointer to a mySetSet struct to which it add member to
 * @param aSet Pointer to Pointer to a mySet struct to add
 */
void setset_insert(mySetSet *aSetSet, mySet **aSet)
{
	if (aSetSet->noOfElt == 0)
	{
		aSetSet->elt = (mySet **) malloc(sizeof(mySet *));
		aSetSet->elt[0] = (*aSet);
	}
	else
	{
		aSetSet->elt = (mySet **) realloc(aSetSet->elt, sizeof(mySet *)*(aSetSet->noOfElt+1));

		aSetSet->elt[aSetSet->noOfElt] = (*aSet);
	}

	aSetSet->noOfElt++;
}

/**
 * It allocate and initialise as zero a new myOrderSet struct.
 * @return Pointer to the new myOrderSet struct
 */
myOrderSet *orderSet_new()
{
	myOrderSet *returnValue;

	returnValue = (myOrderSet *) malloc(sizeof(myOrderSet));
	memset(returnValue, 0, sizeof(myOrderSet));

	return returnValue;
}

/**
 * It make a new copy of a existing myOrderSet struct
 * anOrderSet2. The result copy will be stored in anOrderSet1.
 * It will free up memory of destination myOrderSet struct if
 * it is not empty.
 * <b>Suspected bugs:</b> But it will not free up individual mySet
 * members of original 
 * destination myOrderSet struct.
 * @param anOrderSet1 Pointer to destination myOrderSet struct.
 *                    It must be a instantiated struct
 * @param anOrderSet2 Pointer to source myOrderSet struct.
 */
void orderSet_copy(myOrderSet *anOrderSet1, myOrderSet *anOrderSet2)
{
	int i;

	if (anOrderSet1->noOfElt > 0)
	{
		free(anOrderSet1->dimNo);
		free(anOrderSet1->orderList);
		// don't need to free the details of orderList
	}

	anOrderSet1->noOfElt = anOrderSet2->noOfElt;
	anOrderSet1->dimNo = (int *) malloc(sizeof(int)*anOrderSet1->noOfElt);
	memcpy(anOrderSet1->dimNo, anOrderSet2->dimNo, sizeof(int)*anOrderSet1->noOfElt);
	
	anOrderSet1->orderList = (mySet **) malloc(sizeof(mySet *)*anOrderSet1->noOfElt);
	for (i = 0; i < anOrderSet1->noOfElt; i++)
	{
		anOrderSet1->orderList[i] = set_new();
		set_copy(anOrderSet1->orderList[i], anOrderSet2->orderList[i]);
	}
}

/**
 * It insert a mySet struct with a dimNo as a new member of a
 * myOrderSet struct. It WILL NOT check for duplication.
 * @param anOrderSet The myOrderSet struct to which it add memeber to
 * @param dimNo The dimension of the new mySet struct to insert
 * @param anOrdering The new mySet struct to insert
 */
void orderSet_insert(myOrderSet *anOrderSet, int dimNo, mySet **anOrdering)
{
	int noOfElt;

	noOfElt = anOrderSet->noOfElt;
	if (noOfElt == 0)
	{
		anOrderSet->dimNo = (int *) malloc(sizeof(int));
		anOrderSet->orderList = (mySet **) malloc(sizeof(mySet *));

		anOrderSet->dimNo[0] = dimNo;
		anOrderSet->orderList[0] = (*anOrdering);
	}
	else
	{
		anOrderSet->dimNo = (int *) realloc(anOrderSet->dimNo, sizeof(int)*(noOfElt+1));
		anOrderSet->orderList = (mySet **) realloc(anOrderSet->orderList, sizeof(mySet *)*(noOfElt+1));

		anOrderSet->dimNo[noOfElt] = dimNo;
		anOrderSet->orderList[noOfElt] = (*anOrdering);
	}

	anOrderSet->noOfElt++;
}

/**
 * It update the mySet struct member of a myOrderSet struct having 
 * the specified dimension number. It the member corresponding to
 * the dimension number cannot be found, it is inserted as a new
 * member.
 * @param anOrderSet Pointer to the myOrderSet struct to be updated
 * @param dimNo The dimemsion number of member to be updated
 * @param anOrdering The mySet struct having the updated value
 */
void orderSet_insertAndUpdate(myOrderSet *anOrderSet, int dimNo, mySet **anOrdering)
{
	int noOfElt;

	int i;

	int isFound;
	int index;

	noOfElt = anOrderSet->noOfElt;
	if (noOfElt == 0)
	{
		anOrderSet->dimNo = (int *) malloc(sizeof(int));
		anOrderSet->orderList = (mySet **) malloc(sizeof(mySet *));

		anOrderSet->dimNo[0] = dimNo;
		anOrderSet->orderList[0] = (*anOrdering);

		anOrderSet->noOfElt++;
	}
	else
	{
		// check whether dimNo exists or not
		isFound = 0;
		for (i = 0; i < noOfElt; i++)
		{
			if (anOrderSet->dimNo[i] == dimNo)
			{
				isFound = 1;
				index = i;
				break;
			}
		}

		if (isFound == 1)
		{
			// replace it
			set_free(anOrderSet->orderList[index]);
			free(anOrderSet->orderList[index]);

			anOrderSet->orderList[index] = (*anOrdering);

			// no need to update the noOfElt
		}
		else
		{
			anOrderSet->dimNo = (int *) realloc(anOrderSet->dimNo, sizeof(int)*(noOfElt+1));
			anOrderSet->orderList = (mySet **) realloc(anOrderSet->orderList, sizeof(mySet *)*(noOfElt+1));

			anOrderSet->dimNo[noOfElt] = dimNo;
			anOrderSet->orderList[noOfElt] = (*anOrdering);

			anOrderSet->noOfElt++;
		}
	}

	
}

/**
 * It free up memory occupied by the array holding reference of members
 * of a myOrderSet struct. It does not free up memory of individual
 * members of the myOrderSet itself.
 * @param anOrderSet Pointer to the myOrderSet struct to be free
 */
void orderSet_free(myOrderSet *anOrderSet)
{

	if (anOrderSet->noOfElt > 0)
	{
		//set_free(anOrders

		// just free the array
		free(anOrderSet->dimNo);
		free(anOrderSet->orderList);
	}
}

/**
 * It free up all memory occupied by a myOrderSet struct. It will
 * recursively free all its members. But it does not free up
 * myOrderSet struct itself.
 * @param anOrderSet Pointer to the myOrderSet struct to be free
 */
void orderSet_freeAll(myOrderSet *anOrderSet)
{
	int i;

	if (anOrderSet->noOfElt > 0)
	{
		//set_free(anOrders

		// just free the array
		free(anOrderSet->dimNo);
		for (i = 0; i < anOrderSet->noOfElt; i++)
		{
			set_free(anOrderSet->orderList[i]);
			free(anOrderSet->orderList[i]);
		}

		free(anOrderSet->orderList);
	}
}

/**
 * It allocate  memory to a new myClassSet struct. It will
 * initialise it to zero.
 * @return Pointer to the newly created myClassSet struct
 */
myClassSet *classSet_new()
{
	myClassSet *returnValue;

	returnValue = (myClassSet *) malloc(sizeof(myClassSet));
	memset(returnValue, 0, sizeof(myClassSet));

	return returnValue;
}

/**
 * It insert the given classID into the mySet struct member having the
 * specified dimemsion number. If mySet struct with specified
 * dimemsion number is not found, a new mySet struct is created and
 * added to the myClassSet struct.
 * <b>Warning: Suspected bug was found.</b> Function set_insertNonRedundant()
 * should be used instead.
 * @param aClassSet Pointer to myClassSet struct to be updated
 * @param dimNo Dimension number of the mySet struct to whcih the
 *              classID should add to
 * @param classID The value to be inserted
 */
void classSet_insert(myClassSet *aClassSet, int dimNo, int classID)
{
	int noOfElt;

	int dimNoIndex;
	int i;

	noOfElt = aClassSet->noOfElt;
	if (noOfElt == 0)
	{
		aClassSet->dimNo = (int *) malloc(sizeof(int));
		aClassSet->list = (mySet **) malloc(sizeof(mySet *));

		aClassSet->dimNo[0] = dimNo;
		aClassSet->list[0] = set_new();
		set_insert(aClassSet->list[0], classID);

		aClassSet->noOfElt++;
	}
	else
	{
		// find dimNo index
		dimNoIndex = -1;
		for (i = 0; i < noOfElt; i++)
		{
			if (dimNo == aClassSet->dimNo[i])
			{
				dimNoIndex = i;
				break;
			}
		}

		if (dimNoIndex == -1)
		{
			// have to create an entry for dimNo
			aClassSet->dimNo = (int *) realloc(aClassSet->dimNo, sizeof(int)*(noOfElt+1));
			aClassSet->list = (mySet **) realloc(aClassSet->list, sizeof(mySet *)*(noOfElt+1));
			dimNoIndex = noOfElt;

			aClassSet->dimNo[noOfElt] = dimNo;
			aClassSet->list[noOfElt] = set_new();

			set_insertNonRedundant(aClassSet->list[noOfElt], classID);
			//set_insert(aClassSet->list[noOfElt], classID);

			aClassSet->noOfElt++;
		}
		else
		{
			set_insert(aClassSet->list[dimNoIndex], classID);
		}
		//aClassSet->dimNo[noOfElt] = dimNo;
		//aClassSet->list[noOfElt] = (*anOrdering);
	}

	
}

/**
 * It free up all memory occupied by members of a myClassSet struct. It
 * recursively free up all its members. But it does not free up the
 * myClassSet struct itself.
 * @param aClassSet Pointer to myClassSet struct to free
 */
void classSet_free(myClassSet *aClassSet)
{
	int i;

	if (aClassSet->noOfElt > 0)
	{
		//set_free(anOrders

		// just free the array
		// and need to free aList
		free(aClassSet->dimNo);
		for (i = 0; i < aClassSet->noOfElt; i++)
		{
			set_free(aClassSet->list[i]);
			free(aClassSet->list[i]);
		}
		free(aClassSet->list);
	}
}

/**
 * It perform set union of two myClassSet struct such
 * that it works on individual integer inside mySet.
 * It will check for duplication so that the result
 * will still be a set. The resulting myClassSet struct
 * will be stored in aClassSet1. aClassSet2 is unchanged.
 * <b>Suspected bug:</b>tempSet1 has not been free
 * @param aClassSet1 Pointer to myClassSet struct
 * @param aClassSet2 Pointer to myClassSet struct.
 *                   Unchanged.
 */

void classSet_unionNonRedundant(myClassSet *aClassSet1, myClassSet *aClassSet2)
{
	int i, j;

	int isFound;

	int noOfElt1;
	int noOfElt2;

	mySet **list1;
	mySet **list2;

	int *dimNo1;
	int *dimNo2;

	int *isExist;

	mySet *tempSet1;
	int noOfAdd;
	int finalNoOfElt;

	int eltNo;

	int *finalDimNo;

	int tempNo;

	isFound = 0;

	noOfElt1 = aClassSet1->noOfElt;
	noOfElt2 = aClassSet2->noOfElt;

	list1 = aClassSet1->list;
	list2 = aClassSet2->list;

	dimNo1 = aClassSet1->dimNo;
	dimNo2 = aClassSet2->dimNo;

	isExist = (int *) malloc(sizeof(int)*noOfElt2);
	memset(isExist, 0, sizeof(int)*noOfElt2);

	tempSet1 = set_new();
	for (i = 0; i < noOfElt1; i++)
	{
		set_insert(tempSet1, dimNo1[i]);
	}

	for (i = 0; i < noOfElt2; i++)
	{
		eltNo = dimNo2[i];
		isExist[i] = set_isContain(tempSet1, eltNo);
	}
	noOfAdd = 0;
	for (i = 0; i < noOfElt2; i++)
	{
		if (isExist[i] == 0)
		{
			noOfAdd++;
		}
	}

	finalNoOfElt = noOfElt1+noOfAdd;

	finalDimNo = (int *) malloc(sizeof(int)*finalNoOfElt);
	if (noOfElt1 > 0)
	{
		memcpy(finalDimNo, dimNo1, sizeof(int)*noOfElt1);
	}

	if (noOfAdd > 0)
	{
		list1 = (mySet **) realloc(list1, sizeof(mySet *)*(finalNoOfElt));
	}

	tempNo = 0;
	for (i = 0; i < noOfElt2; i++)
	{
		if (isExist[i] == 0)
		{
			eltNo = dimNo2[i];

			finalDimNo[tempNo+noOfElt1] = eltNo;


			list1[tempNo+noOfElt1] = set_new();

			set_copy(list1[tempNo+noOfElt1], list2[i]);

			//

			tempNo++;
		}
		else
		{
			for (j = 0; j < noOfElt1; j++)
			{
				if (dimNo1[j] == dimNo2[i])
				{
					set_unionNonRedundant(list1[j], list2[i]);
					break;
				}
			}
		}
	}

	free(isExist);

	free(dimNo1);
	// no need to free list1

	aClassSet1->dimNo = finalDimNo;
	aClassSet1->noOfElt = finalNoOfElt;
	aClassSet1->list = list1;

}

/**
 * It perform set intersaction on two myClassSet struct.
 * Only mySet struct having same dimension numer in both
 * of original myClassSet struct will be retained. In
 * each retained mySet struct, intersaction will also be
 * performed.
 * <b>Suspected bugs:</b>The memory of list1 and all its
 * member has not been free
 * @param aClassSet1
 * @param aClassSet2
 */
void classSet_intersection(myClassSet *aClassSet1, myClassSet *aClassSet2)
{
	int noOfElt1, noOfElt2;
	int *dimNo1;
	int *dimNo2;
	mySet **list1;
	mySet **list2;

	int *finalNoOfElt;

	int *elt1;
	int *elt2;

	//int *finalDimNo;
	//mySet **finalList;

	int **isMark;

	//mySet *eltNo1;
	//mySet *eltNo2;

	//int *tempNo;

	int indexFound;

	int tempCount;

	int count;

	int i, j, k, l;

	dimNo1 = aClassSet1->dimNo;
	list1 = aClassSet1->list;
	noOfElt1 = aClassSet1->noOfElt;

	dimNo2 = aClassSet2->dimNo;
	list2 = aClassSet2->list;
	noOfElt2 = aClassSet2->noOfElt;

	if (noOfElt1 > 0)
	{
		finalNoOfElt = (int *) malloc(sizeof(int)*noOfElt1);
		memset(finalNoOfElt, 0, sizeof(int)*noOfElt1);

		isMark = (int **) malloc(sizeof(int *)*noOfElt1);
		memset(isMark, 0, sizeof(int *)*noOfElt1);

		for (i = 0; i < noOfElt1; i++)
		{
			isMark[i] = (int *) malloc(sizeof(int)*list1[i]->noOfElt);
			memset(isMark[i], 0, sizeof(int)*list1[i]->noOfElt);

			indexFound = -1;
			for (j = 0; j < noOfElt2; j++)
			{
				if (dimNo1[i] == dimNo2[j])
				{
					indexFound = j;
					break;
				}
			}

			if (indexFound != -1)
			{
				elt1 = list1[i]->elt;
				elt2 = list2[indexFound]->elt;
				for (k = 0; k < list1[i]->noOfElt; k++)
				{
					for (l = 0; l < list2[indexFound]->noOfElt; l++)
					{
						if (elt1[k] == elt2[l])
						{
							isMark[i][k] = 1;
							finalNoOfElt[i]++;
							break;
						}
					}
				}
			}
			else
			{
				// do nothing
			}
		}

		



		count = 0;
		for (i = 0; i < noOfElt1; i++)
		{
			if (finalNoOfElt[i] != 0)
			{
				count++;
			}
		}

		aClassSet1->noOfElt = count;
		aClassSet1->dimNo = (int *) malloc(sizeof(int)*count);
		aClassSet1->list = (mySet **) malloc(sizeof(mySet *)*count);

				//finalElt = (int **) malloc(sizeof(int *)*noOfElt1);

		tempCount = 0;
		for (i = 0; i < noOfElt1; i++)
		{
			if (finalNoOfElt[i] != 0)
			{
				aClassSet1->dimNo[tempCount] = dimNo1[i];

				//finalElt[i] = (int *) malloc(sizeof(int)*finalNoOfElt[i]);
				aClassSet1->list[tempCount] = set_new();


				elt1 = list1[i]->elt;
				for (k = 0; k < list1[i]->noOfElt; k++)
				{
					if (isMark[i][k] == 1)
					{
						//finalElt[i][tempNo[i]] = elt1[k];
						//tempNo[i]++;
						set_insert(aClassSet1->list[tempCount], elt1[k]);
					}
				}

				tempCount++;
			}
		}

		free(dimNo1);
		free(elt1);


		free(finalNoOfElt);
		for (i = 0; i < noOfElt1; i++)
		{
			free(isMark[i]);
		}
		free(isMark);
	}
}

/**
 * It make a copy of myClassSet struct. The resulting
 * struct will be pointed by aClassSet1.
 * @param aClassSet1 Pointer to destination myClassSet
 * struct. It must be a valid struct(memory already allocated).
 * @param aClassSet2 Pointer to source myClassSet
 * struct. It remain unchanged after operation.
 */
void classSet_copy(myClassSet *aClassSet1, myClassSet *aClassSet2)
{
	int i;

	if (aClassSet1->noOfElt > 0)
	{
		free(aClassSet1->dimNo);
		for (i = 0; i < aClassSet1->noOfElt; i++)
		{
			set_free(aClassSet1->list[i]);
			free(aClassSet1->list[i]);
		}
		free(aClassSet1->list);
	}

	aClassSet1->noOfElt = aClassSet2->noOfElt;
	aClassSet1->dimNo = (int *) malloc(sizeof(int)*aClassSet1->noOfElt);
	memcpy(aClassSet1->dimNo, aClassSet2->dimNo, sizeof(int)*aClassSet1->noOfElt);

	aClassSet1->list = (mySet **) malloc(sizeof(mySet *)*aClassSet1->noOfElt);
	for (i = 0; i < aClassSet1->noOfElt; i++)
	{
		aClassSet1->list[i] = set_new();
		set_copy(aClassSet1->list[i], aClassSet2->list[i]);
	}
}

/**
 * It look for classID in mySet struct members of myClassSet
 * struct having a specified dimension number.
 * @param aClassSet Pointer to myClassSet struct in
 * whcih we look for value
 * @param dimNo Dimension number the classID correspond
 * to
 * @param classID The target value we look for
 */
int classSet_isContain(myClassSet *aClassSet, int dimNo, int classID)
{
	int returnValue;

	mySet *tempSet;

	int *elt;
	int i, j;

	int noOfElt;

	returnValue = 0;
	for (i = 0; i < aClassSet->noOfElt; i++)
	{
		if (aClassSet->dimNo[i] == dimNo)
		{
			// do the further search
			tempSet = aClassSet->list[i];

			elt = tempSet->elt;
			noOfElt = tempSet->noOfElt;

			for (j = 0; j < noOfElt; j++)
			{
				if (elt[j] == classID)
				{
					returnValue = 1;
					break;
				}
			}
			break;
		}
	}

	return returnValue;
}

/**
 * It check if a myClassSetSet struct contains a member
 * having same value as a myClassSet struct.
 * @param aClassSetSet Pointer to myClassSetSet struct
 *                     that we look into
 * @param aClassSet Pointer to myClassSet struct that 
 *                 we look for
 * @return 0 if not found. 1 if found
 */
int classSetSet_isContain_set(myClassSetSet *aClassSetSet, myClassSet *aClassSet)
{
	int returnValue;

	//mySet *tempSet;

	int i; //, j, k;

	int noOfElt;
	//int checkDimNo;

	myClassSet **elt;
	myClassSet *eltClassSet;

	//int eltID;
	//int checkID;

	int isContain;

	noOfElt = aClassSetSet->noOfElt;
	elt = aClassSetSet->elt;

	returnValue = 0;

	if (aClassSet->noOfElt > 0)
	{
	
		//checkDimNo = aClassSet->dimNo[0];
		for (i = 0; i < noOfElt; i++)
		{
			eltClassSet = elt[i];

			isContain = classSet_isContainAll(eltClassSet, aClassSet);
			
			if (isContain == 1)
			{
				returnValue = 1;
				break;
			}
		}
	}

	return returnValue;
}

/**
 * It check if all members of aClassSet1 can be found 
 * in aClassSet2. Each member(mySet struct) of myClassSet struct has
 * only 2 elements.
 * @param aClassSet1 Pointer to myClassSet to look for
 * @param aClassSet2 Pointer to myClassSet to look into
 * @return 0 if false. 1 if true.
 */
int classSet_isEqualOrIsWeaker(myClassSet *aClassSet1, myClassSet *aClassSet2)
{
	int returnValue;

	int noOfElt1, noOfElt2;

	mySet **list1;
	mySet **list2;

	int i, j;

	int elt11, elt12, elt21, elt22;

	int *dimNo1;
	int *dimNo2;

	int isExist;


	noOfElt1 = aClassSet1->noOfElt;
	noOfElt2 = aClassSet2->noOfElt;

	list1 = aClassSet1->list;
	list2 = aClassSet2->list;

	dimNo1 = aClassSet1->dimNo;
	dimNo2 = aClassSet2->dimNo;

	if ((noOfElt1 == noOfElt2) || (noOfElt1 < noOfElt2))
	{
		// if the same
		returnValue = 1;
		for (i = 0; i < noOfElt1; i++)
		{
			elt11 = list1[i]->elt[0];
			elt12 = list1[i]->elt[1];

			isExist = 0;
			for (j = 0; j < noOfElt2; j++)
			{
				// if dimNo is equal
				if (dimNo1[i] == dimNo2[j])
				{
					elt21 = list2[j]->elt[0];
					elt22 = list2[j]->elt[1];

					if ((elt11 == elt21) && (elt12 == elt22))
					{
						isExist = 1;
						break;
					}
				}
			}

			if (isExist == 1)
			{
				// do nothing
			}
			else
			{
				// not weaker
				returnValue = 0;
				break;
			}
		}
	}
	else
	{
		// not weaker
		returnValue = 0;

	}

	return returnValue;
}

/**
 * It check if aClassSet2 is stronger than aClassSet1.
 * A is stronger than B if all relation in B can be found
 * in A and A has more relation than B.
 * <b>Suspected bug:</b>It will return 0 if some elements
 * in aClassSet1 cannot be found in aClassSet2.
 * <b>Assumption:</b>Each member(mySet struct) of myClassSet
 * struct has only 2 elements.
 * @param aClassSet1
 * @param aClassSet2
 * @return 0 if aClassSet2 is not stronger than aClassSet1.
 *         1 if aClassSet2 is stronger than aClassSet1
 */
int classSet_isStronger(myClassSet *aClassSet1, myClassSet *aClassSet2)
{
	int returnValue;

	int noOfElt1, noOfElt2;

	mySet **list1;
	mySet **list2;

	int i, j;

	int elt11, elt12, elt21, elt22;

	int *dimNo1;
	int *dimNo2;

	int isExist;


	noOfElt2 = aClassSet1->noOfElt;
	noOfElt1 = aClassSet2->noOfElt;

	list2 = aClassSet1->list;
	list1 = aClassSet2->list;

	dimNo2 = aClassSet1->dimNo;
	dimNo1 = aClassSet2->dimNo;

	//if ((noOfElt1 == noOfElt2) || (noOfElt1 < noOfElt2))
	if (noOfElt1 < noOfElt2)
	{
		// if the same
		returnValue = 1;
		for (i = 0; i < noOfElt1; i++)
		{
			elt11 = list1[i]->elt[0];
			elt12 = list1[i]->elt[1];

			isExist = 0;
			for (j = 0; j < noOfElt2; j++)
			{
				// if dimNo is equal
				if (dimNo1[i] == dimNo2[j])
				{
					elt21 = list2[j]->elt[0];
					elt22 = list2[j]->elt[1];

					if ((elt11 == elt21) && (elt12 == elt22))
					{
						isExist = 1;
						break;
					}
				}
			}

			if (isExist == 1)
			{
				// do nothing
			}
			else
			{
				// not weaker
				returnValue = 0;
				break;
			}
		}
	}
	else
	{
		// not weaker
		returnValue = 0;

	}

	return returnValue;
}

/**
 * It check if ANY member of a myClassSetSet struct
 * is weaker or equal to a myClassSet struct
 * <b>Warning:</b>The purpose of this functoin
 * is quite strange!!!
 * @param aClassSetSet Contain members of mySet struct
 * @param aClassSet Contains mySet struct of different
 *                  dimension
 * @return 0 if none of the member of aClassSetSet is
 * weaker or equal to aClassSet. 1 if at least one member
 * of aClassSetSet is weaker or equal to aClassSet.
 */
int classSetSet_isContainOrIsWeaker(myClassSetSet *aClassSetSet, myClassSet *aClassSet)
{
	int returnValue;

	int noOfElt;
	int i; //, j;

	int isEqualOrIsWeaker;

	myClassSet **elt;

	noOfElt = aClassSetSet->noOfElt;
	elt = aClassSetSet->elt;

	returnValue = 0;
	for (i = 0; i < noOfElt; i++)
	{
		isEqualOrIsWeaker = classSet_isEqualOrIsWeaker(elt[i], aClassSet);

		if (isEqualOrIsWeaker == 1)
		{
			returnValue = 1;
			break;
		}
	}

	return returnValue;
}

/**
 * It delete index-th element of myClassSetSet struct.
 * <b>Suspect bug:</b>The myClassSet struct being deleted
 * has not been free
 * @param aClassSetSet myClassSetSet struct containing from
 * which we delete item from
 * @param index The index of the member of aClassSetSet to
 * deleted 
 */
void classSetSet_removeIndex(myClassSetSet *aClassSetSet, int index)
{
	int noOfElt;

	myClassSet **elt;
	myClassSet **newElt;

	noOfElt = aClassSetSet->noOfElt;
	elt = aClassSetSet->elt;

	if (noOfElt > 0)
	{
		newElt = (myClassSet **) malloc(sizeof(myClassSet *)*(noOfElt-1));
		if (index > 0)
		{
			memcpy(newElt, elt, sizeof(myClassSet *)*index);
		}

		if (index < noOfElt-1)
		{
			memcpy(&(newElt[index]), &(elt[index+1]), sizeof(myClassSet *)*(noOfElt-1-index));
		}
		aClassSetSet->elt = newElt;
		aClassSetSet->noOfElt--;

		free(elt);
	}
}

/**
 * <b>Suspected bug:</b>If we remove element from
 * aClassSetSet inside for loop, loop counter i will
 * point to wrong position.
 */
void classSetSet_removeStronger(myClassSetSet *aClassSetSet, myClassSet *aClassSet)
{
	int noOfElt;
	int i; //, j;

	int isStronger;

	myClassSet **elt;

	noOfElt = aClassSetSet->noOfElt;
	

	//*** NEW
	//for (i = 0; i < noOfElt; i++)
	for (i = noOfElt-1; i >= 0; i--)
	{
		elt = aClassSetSet->elt;

		isStronger = classSet_isStronger(elt[i], aClassSet);

		if (isStronger == 1)
		{
			// remove index i
			classSetSet_removeIndex(aClassSetSet, i);
		}
	}
}

/**
 * It check if aClassSet contains all members of 
 * smallClassSet. <b>Assumption:</b>All member
 * of smallClassSet has only one element.
 * @param aClassSet Its member can have more than
 * one elelment.
 * @param smallClassSet We search for it in aClassSet.
 */
int classSet_isContainAll(myClassSet *aClassSet, myClassSet *smallClassSet)
{
	int returnValue;

	mySet *tempSet;

	int *elt;
	int i, j, k;

	int dimNo;
	int classID;

	int isFound;

	int noOfElt;

	returnValue = 1;
	for (k = 0; k < smallClassSet->noOfElt; k++)
	{
		dimNo = smallClassSet->dimNo[k];
		classID = smallClassSet->list[k]->elt[0];

		isFound = 0;
		for (i = 0; i < aClassSet->noOfElt; i++)
		{
			if (aClassSet->dimNo[i] == dimNo)
			{
				// do the further search
				tempSet = aClassSet->list[i];

				elt = tempSet->elt;
				noOfElt = tempSet->noOfElt;

				for (j = 0; j < noOfElt; j++)
				{
					if (elt[j] == classID)
					{
						//returnValue = 1;
						isFound = 1;
						break;
					}
				}
				break;
			}
		}

		if (isFound == 0)
		{
			returnValue = 0;
			break;
		}
	}

	return returnValue;
}

/**
 * It allocate and set zero memory for a new myClassSetSet
 * struct
 * @return Pointer to the newly created myClassSetSet struct
 */
myClassSetSet *classSetSet_new()
{
	myClassSetSet *returnValue;

	returnValue = (myClassSetSet *) malloc(sizeof(myClassSetSet));
	memset(returnValue, 0, sizeof(myClassSetSet));

	return returnValue;
}

/**
 * It make a new copy of aClassSetSet2. The result is point
 * to by aClassSetSet1. It does not clone individual member
 * of aClassSetSet2.
 * <b>Suspected bug:</b>It does not free memory occupied by
 * individual member of aClassSetSet1 if it is not empty.
 * @param aClassSetSet1 Pointer to the new copy of
 * myClassSetSet struct. It must be a valid myClassSetSet
 * struct as a argument.
 * @param aClassSetSet2 What we want to copy
 */
void classSetSet_clone(myClassSetSet *aClassSetSet1, myClassSetSet *aClassSetSet2)
{
	myClassSet **oldElt;
	
	oldElt = aClassSetSet1->elt;
	if (aClassSetSet1->noOfElt > 0)
	{
		free(oldElt);
	}
	aClassSetSet1->elt = (myClassSet **) malloc(sizeof(myClassSet *)*aClassSetSet2->noOfElt);

	memcpy(aClassSetSet1->elt, aClassSetSet2->elt, sizeof(myClassSet *)*aClassSetSet2->noOfElt);


	aClassSetSet1->noOfElt = aClassSetSet2->noOfElt;
}

/**
 * It insert aClassSet to aClassSetSet as a new member.
 * It does not check for duplication of member.
 * @param aClassSetSet Where we want to inset to
 * @param aClassSet What we want to insert
 */
void classSetSet_insert(myClassSetSet *aClassSetSet, myClassSet *aClassSet)
{
	//myClassSet *anElt;
	int noOfElt;

	noOfElt = aClassSetSet->noOfElt;

	if (noOfElt == 0)
	{
		aClassSetSet->elt = (myClassSet **) malloc(sizeof(myClassSet *));
		aClassSetSet->elt[0] = aClassSet;
	}
	else
	{
		aClassSetSet->elt = (myClassSet **) realloc(aClassSetSet->elt, sizeof(myClassSet *)*(noOfElt+1));
		aClassSetSet->elt[noOfElt] = aClassSet;
	}

	aClassSetSet->noOfElt++;
}

/**
 * It insert aClassSet to aClassSetSet as a new member.
 * It will check for duplication of member.
 * <b>Suspected bug:</b>aClassSetSet's member can only
 * have one element.
 * @param aClassSetSet Where we want to inset to
 * @param aClassSet What we want to insert
 * @return 1 if no insertion is done. 0 if insertion
 * is done.
 */
int classSetSet_insertNonRedundant(myClassSetSet *aClassSetSet, myClassSet **aClassSet)
{
	//myClassSet *anElt;
	int noOfElt;

	int returnValue;

	//int isFound;

	int isContain;
	myClassSet **elt;

	//int i;

	noOfElt = aClassSetSet->noOfElt;
	elt = aClassSetSet->elt;

	if (noOfElt == 0)
	{
		aClassSetSet->elt = (myClassSet **) malloc(sizeof(myClassSet *));
		aClassSetSet->elt[0] = (*aClassSet);

		aClassSetSet->noOfElt++;

		returnValue = 0;
	}
	else
	{
		// found
		isContain = classSetSet_isContain_set(aClassSetSet, (*aClassSet));


		if (isContain == 1)
		{
			// no need to insert

			returnValue = 1;
		}
		else
		{
			aClassSetSet->elt = (myClassSet **) realloc(aClassSetSet->elt, sizeof(myClassSet *)*(noOfElt+1));
			aClassSetSet->elt[noOfElt] = (*aClassSet);

			aClassSetSet->noOfElt++;

			returnValue = 0;
		}
	}

	return returnValue;
}

/**
 * It recursively free memory of all members of
 * aClassSetSet. It does not free myClassSetSet
 * struct itself.
 * @param aClassSetSet What we want to free from
 * memory.
 */
void classSetSet_free(myClassSetSet *aClassSetSet)
{
	int noOfElt;
	int i;

	myClassSet **elt;

	noOfElt = aClassSetSet->noOfElt;
	elt = aClassSetSet->elt;
	for (i = 0; i < noOfElt; i++)
	{
		classSet_free(elt[i]);
		free(elt[i]);
	}
	free(elt);
}

/**
 * It allocate and set zero memory to a new myClassSetSetSet
 * struct.
 * @return Pointer to the new myClassSetSetSet struct
 */
myClassSetSetSet *classSetSetSet_new()
{
	myClassSetSetSet *returnValue;

	returnValue = (myClassSetSetSet *) malloc(sizeof(myClassSetSetSet));
	memset(returnValue, 0, sizeof(myClassSetSetSet));

	return returnValue;
}

/**
 * It insert aClassSetSet as a new element to
 * aClassSetSetSet. It does not check for duplication.
 * @param aClassSetSetSet To which we add element to.
 * @param pointID Identify each member of aClassSetSetSet
 * @param aClassSetSet Element to add to aClassSetSetSet
 */
void classSetSetSet_insert(myClassSetSetSet *aClassSetSetSet, int pointID, myClassSetSet **aClassSetSet)
{
	int noOfElt;

	noOfElt = aClassSetSetSet->noOfElt;

	if (noOfElt == 0)
	{
		aClassSetSetSet->elt = (myClassSetSet **) malloc(sizeof(myClassSetSet *));
		aClassSetSetSet->elt[0] = (*aClassSetSet);

		aClassSetSetSet->pointID = (int *) malloc(sizeof(int));
		aClassSetSetSet->pointID[0] = pointID;
	}
	else
	{
		aClassSetSetSet->elt = (myClassSetSet **) realloc(aClassSetSetSet->elt, sizeof(myClassSetSet *)*(noOfElt+1));
		aClassSetSetSet->elt[noOfElt] = (*aClassSetSet);

		aClassSetSetSet->pointID = (int *) realloc(aClassSetSetSet->pointID, sizeof(int)*(noOfElt+1));
		aClassSetSetSet->pointID[noOfElt] = pointID;
	}

	aClassSetSetSet->noOfElt++;
}

/**
 * It free up memory of array that used to hold
 * reference of members of aClassSetSetSet. It does
 * not free recursively each individual myClassSetSet
 * members. It also does not free itself.
 * @param aClassSetSetSet What we want to free
 */
void classSetSetSet_free(myClassSetSetSet *aClassSetSetSet)
{
	if (aClassSetSetSet->noOfElt > 0)
	{
		free(aClassSetSetSet->pointID);
		free(aClassSetSetSet->elt);
	}
}

/**
 * It free up memory used by all members of aSet.
 * It will set number of element to 0. It will not 
 * free itself.
 * @param aSet The mySet struct we want to empty
 */
void set_empty(mySet *aSet)
{
	if (aSet->noOfElt > 0)
	{
		free(aSet->elt);
		aSet->noOfElt = 0;
		aSet->elt = NULL;
	}
}

/**
 * It insert a new element into aClassSet without checking
 * for duplication.
 * @param aClassSet To which we add element
 * @param dimNo The corresponding dimension number of the
 * new element
 * @param aSet The element we want to add
 * @author Sing
 * @since 18 Oct 2006
 */
void classSet_insertSet(myClassSet *aClassSet, int dimNo, mySet *aSet)
{
	int noOfElt;

	noOfElt = aClassSet->noOfElt;
	if (noOfElt == 0)
	{
		aClassSet->dimNo = (int *) malloc(sizeof(int));
		aClassSet->list = (mySet **) malloc(sizeof(mySet *));

		aClassSet->dimNo[0] = dimNo;
		aClassSet->list[0] = aSet;
	}
	else
	{
		aClassSet->dimNo = (int *) realloc(aClassSet->dimNo, sizeof(int)*(noOfElt+1));
		aClassSet->list = (mySet **) realloc(aClassSet->list, sizeof(mySet *)*(noOfElt+1));

		aClassSet->dimNo[noOfElt] = dimNo;
		aClassSet->list[noOfElt] = aSet;
	}
	aClassSet->noOfElt++;
}

/**
 * It union two myClassSet struct such that it work
 * on entire mySet struct. The result will be stored
 * in aClassSet1.
 * @param aClassSet1
 * @param aClassSet2
 * @author Sing
 * @since 18 Oct 2006
 */
void classSet_unionNonRedundant2(myClassSet *aClassSet1, myClassSet *aClassSet2)
{
	int i = 0;
	int j = 0;
	int found = 0;
	for (i=0; i < aClassSet2->noOfElt; i++)
	{
		found = 0;
		j = 0;
		//Looping each member of aClassSet1 to see if mySet exist in aClassSet1
		while (j < aClassSet1->noOfElt && found==0)
		{
			if (aClassSet2->dimNo[i]==aClassSet1->dimNo[j])
			{
				found = set_isEqual(aClassSet2->list[i], aClassSet1->list[j]);
			}
			j++;
		}
		//Directly add the mySet to aClassSet1 if not found
		if (!found)
		{
			classSet_insertSet(aClassSet1, aClassSet2->dimNo[i],
					aClassSet2->list[i]);
		}
	}
}

/**
 * It check if members and their order in aSet1 is exactly the same as
 * in aSet2.
 * @param aSet1
 * @param aSet2
 * @return 0 if ther are not equal. 1 if they are equal.
 * @author Sing
 * @since 18 Oct 2006
 */
int set_isEqual(mySet *aSet1, mySet *aSet2)
{
	int i = 0;
	int equal = 0;
	if (aSet1->noOfElt==aSet2->noOfElt)
	{
		equal = 1;
		while (i < aSet1->noOfElt && equal==1)
		{
			if (aSet1->elt[i] != aSet2->elt[i]) {
				equal = 0;
			}
			i++;
		}
	}
	return equal;
}

/**
 * It output all values of aClassSet to file pointed by filePtr
 * @param filePtr The file to output to
 * @param aClassSet What we want to watch
 * @author Sing
 * @since 18 Oct 2006
 */
void classSet_print(FILE *filePtr, myClassSet* aClassSet)
{
	int i=0;
	fputs("myClass details:\n", filePtr);
	fputs("----------------\n", filePtr);
	for (i = 0; i < aClassSet->noOfElt; i++)
	{
		fprintf(filePtr, "Dim:%d -- ", aClassSet->dimNo[i]);
		set_print(filePtr, aClassSet->list[i]);
		fputs("\n", filePtr);
	}
}

/**
 * <b>Assumption:</b>Each member(mySet) of aClassSet only has 2 elements.
 * We compare by lexicographic order: (1)dimension, (2)classBefore,
 * (3)classAfter e.g. (A1, A2)<(B2, B3)
 * @param aClassSet1
 * @param index1 Which refinement in aClassSet1 we want to compare
 * @param aClassSet2
 * @param index2 Which refinement in aClassSet2 we want to compare
 * @return 1 if aClassSet1 is less than or equal to aClassSet2, 0 otherwise
 * @author Terry
 * @since 21 Oct 2006
 */
int classSet_isLessThanOrEqual(myClassSet* aClassSet1, int index1,
		myClassSet* aClassSet2, int index2)
{
	int isLessThan = 1;
	if (aClassSet2->dimNo[index2] < aClassSet1->dimNo[index1])
	{
		isLessThan = 0;
	}
	else if (aClassSet2->dimNo[index2] == aClassSet1->dimNo[index1])
	{
		if (aClassSet2->list[index2]->elt[0] < aClassSet1->list[index1]->elt[0])
		{
			isLessThan = 0;
		}
		else if (aClassSet2->list[index2]->elt[0] == aClassSet1->list[index1]->elt[0])
		{
			if (aClassSet2->list[index2]->elt[1] < aClassSet1->list[index1]->elt[1])
			{
				isLessThan = 0;
			}
		}
	}
	return isLessThan;
}

mySet *setSort_merage(int mode, mySet *set1, mySet *set2)
{
	mySet *returnValue;

	int index1, index2;
	int value1, value2;

	int noOfElt1;
	int *elt1;

	int noOfElt2;
	int *elt2;

	returnValue = set_new();

	index1 = 0;
	index2 = 0;

	noOfElt1 = set1->noOfElt;
	elt1 = set1->elt;

	noOfElt2 = set2->noOfElt;
	elt2 = set2->elt;

	while (index1 < noOfElt1 && index2 < noOfElt2)
	{
		value1 = elt1[index1];
		value2 = elt2[index2];

		if (value1 == value2)
		{
			set_insert(returnValue, value1);
			index1++;
			index2++;
		}
		else if (value1 < value2)
		{
			if (mode)
			{
				set_insert(returnValue, value1);
			}
			index1++;
		}
		else
		{
			if (mode)
			{
				set_insert(returnValue, value2);
			}
			index2++;
		}
	}
	if (mode)
	{
		while (index1 < noOfElt1)
		{
			value1 = elt1[index1];
			set_insert(returnValue, value1);
			index1++;
		}
		while (index2 < noOfElt2)
		{
			value2 = elt2[index2];
			set_insert(returnValue, value2);
			index2++;
		}
	}

	return returnValue;
}

mySet *setSort_merage_bulky(int mode, mySet *set1, mySet *set2)
{
	mySet *returnValue;

	int index1, index2;
	int value1, value2;

	int noOfElt1;
	int *elt1;

	int noOfElt2;
	int *elt2;

	int count;

	returnValue = set_new();

	index1 = 0;
	index2 = 0;

	noOfElt1 = set1->noOfElt;
	elt1 = set1->elt;

	noOfElt2 = set2->noOfElt;
	elt2 = set2->elt;

	returnValue->elt = (int *) malloc(sizeof(int)*(noOfElt1+noOfElt2));
	count = 0;
	while (index1 < noOfElt1 && index2 < noOfElt2)
	{
		value1 = elt1[index1];
		value2 = elt2[index2];

		if (value1 == value2)
		{
			//set_insert(returnValue, value1);
			returnValue->elt[count] = value1;
			count++;
			index1++;
			index2++;
		}
		else if (value1 < value2)
		{
			if (mode)
			{
				//set_insert(returnValue, value1);
				returnValue->elt[count] = value1;
				count++;
			}
			index1++;
		}
		else
		{
			if (mode)
			{
				//set_insert(returnValue, value2);
				returnValue->elt[count] = value2;
				count++;
			}
			index2++;
		}
	}
	if (mode)
	{
		while (index1 < noOfElt1)
		{
			value1 = elt1[index1];
			//set_insert(returnValue, value1);
			returnValue->elt[count] = value1;
			count++;
			index1++;
		}
		while (index2 < noOfElt2)
		{
			value2 = elt2[index2];
			//set_insert(returnValue, value2);
			returnValue->elt[count] = value2;
			count++;
			index2++;
		}
	}
	returnValue->noOfElt = count;

	return returnValue;
}


// require the set is sorted
mySet *setSort_union(mySet *set1, mySet *set2)
{
	mySet *returnValue;

	returnValue = setSort_merage(1, set1, set2);
	
	return returnValue;
}

mySet *setSort_union_bulky(mySet *set1, mySet *set2)
{
	mySet *returnValue;

	returnValue = setSort_merage_bulky(1, set1, set2);
	
	return returnValue;
}

// require the set is sorted
mySet *setSort_intersect(mySet *set1, mySet *set2)
{
	mySet *returnValue;

	returnValue = setSort_merage(0, set1, set2);
	
	return returnValue;
}

// require the set is sorted
mySet *setSort_minus(mySet *set1, mySet *set2)
{
	mySet *returnValue;

	int index1, index2;
	int value1, value2;

	int noOfElt1;
	int *elt1;

	int noOfElt2;
	int *elt2;

	returnValue = set_new();

	noOfElt1 = set1->noOfElt;
	elt1 = set1->elt;

	noOfElt2 = set2->noOfElt;
	elt2 = set2->elt;

	index1 = 0;
	index2 = 0;

	while (index1 < noOfElt1 && index2 < noOfElt2)
	{
		value1 = elt1[index1];
		value2 = elt2[index2];

		if (value1 == value2)
		{
			index1++;
			index2++;
		}
		else if (value1 < value2)
		{
			set_insert(returnValue, value1);
			index1++;
		}
		else
		{
			index2++;
		}
	}
	while (index1 < noOfElt1)
	{
		value1 = elt1[index1];
		set_insert(returnValue, value1);
		index1++;
	}

	return returnValue;
}

void setSort_insertSort(mySet *set1, int value)
{
	int noOfElt;
	int *elt;
	int *new_elt;

	int i;
	int index;
	int insertIndex;

	noOfElt = set1->noOfElt;
	elt = set1->elt;

	if (noOfElt == 0)
	{
		new_elt = (int *) malloc(sizeof(int)*1);
		
		new_elt[0] = value;
	}
	else
	{
		
		index = -1;
		for (i = noOfElt-1 ; i >= 0 ; i--)
		{
			if (elt[i] <= value)
			{
				index = i;
				break;
			}
		}

		new_elt = (int *) malloc(sizeof(int)*(noOfElt+1));

		insertIndex = index+1;
		new_elt[insertIndex] = value;
		if (insertIndex > 0)
		{
			memcpy(new_elt, elt, sizeof(int)*insertIndex);
		}
		if (noOfElt - insertIndex > 0)
		{
			memcpy(&(new_elt[insertIndex+1]), &(elt[insertIndex]), sizeof(int)*(noOfElt-insertIndex));
		}

		if (noOfElt > 0)
		{
			free(elt);
		}
	}
	
	set1->elt = new_elt;
	set1->noOfElt++;
}

void bitwise_getWordNoAndOffset(int index, int *wordNo, int *offset)
{
	(*wordNo) = index/WORDSIZE;
	(*offset) = index%WORDSIZE;
}

void bitwise_setOne(int *array, int wordNo, int offset)
{
	array[wordNo] |= (1 << offset);
}

void bitwise_setZero(int *array, int wordNo, int offset)
{
	array[wordNo] &= ~(1 << offset);
}

// aSet: stores all "indexes" (or bit-position) to be removed
void bitwise_minus(int *array, mySet *aSet)
{
	int wordNo;
	int offset;

	int index;

	int noOfElt;
	int *elt;

	int i;

	noOfElt = aSet->noOfElt;
	elt = aSet->elt;

	for (i = 0; i < noOfElt; i++)
	{
		index = elt[i];

		bitwise_getWordNoAndOffset(index, &wordNo, &offset);
		bitwise_setZero(array, wordNo, offset);
	}
}

// aSet: stores all "indexes" (or bit-position) to be added
void bitwise_union(int *array, mySet *aSet)
{
	int wordNo;
	int offset;

	int index;

	int noOfElt;
	int *elt;

	int i;

	noOfElt = aSet->noOfElt;
	elt = aSet->elt;

	for (i = 0; i < noOfElt; i++)
	{
		index = elt[i];

		bitwise_getWordNoAndOffset(index, &wordNo, &offset);
		bitwise_setOne(array, wordNo, offset);
	}
}

